/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editpart.policy;

import static org.fortiss.tooling.base.ui.utils.DragAndDropBaseUtils.createConnectionDropContext;

import org.eclipse.draw2d.FigureCanvas;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.gef.EditPart;
import org.eclipse.gef.GraphicalEditPart;
import org.eclipse.gef.Request;
import org.eclipse.gef.commands.Command;
import org.eclipse.gef.editpolicies.LayoutEditPolicy;
import org.eclipse.gef.requests.CreateRequest;
import org.eclipse.swt.widgets.Control;
import org.fortiss.tooling.base.dnd.ElementDropContext;
import org.fortiss.tooling.base.ui.editpart.DiagramEditPartBase;
import org.fortiss.tooling.base.ui.editpart.request.CreateConnectionCompositionRequest;
import org.fortiss.tooling.base.ui.editpart.request.CreateElementCompositionRequest;
import org.fortiss.tooling.base.ui.utils.ZoomUIUtils;
import org.fortiss.tooling.base.utils.LayoutModelElementFactory;
import org.fortiss.tooling.kernel.extension.data.IConnectionCompositionContext;
import org.fortiss.tooling.kernel.extension.data.IElementCompositionContext;
import org.fortiss.tooling.kernel.service.IConnectionCompositorService;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.ui.dnd.ConnectionCompositionSource;
import org.fortiss.tooling.kernel.ui.dnd.ElementCompositionSource;

/**
 * Base class for edit policies relaying create commands to {@link IElementCompositorService}.
 * 
 * @author hoelzl
 */
public abstract class RelayCreateCommandEditPolicyBase extends LayoutEditPolicy {

	/** {@inheritDoc} */
	@Override
	public Command getCommand(Request request) {
		if(request instanceof CreateElementCompositionRequest ||
				request instanceof CreateConnectionCompositionRequest) {
			return getCreateCompositionCommand((CreateRequest)request);
		}
		return super.getCommand(request);
	}

	/** {@inheritDoc} */
	@Override
	protected Command getCreateCommand(CreateRequest request) {
		// this is currently ignored
		return null;
	}

	/**
	 * Creates the command for {@link CreateElementCompositionRequest}s and
	 * {@link CreateConnectionCompositionRequest}s.
	 */
	private Command getCreateCompositionCommand(CreateRequest request) {
		// the request should only contain a position (e.g. mouse clicked not
		// dragged)
		if(request.getSize() != null) {
			return null;
		}
		if(!(getHost().getModel() instanceof EObject)) {
			return null;
		}
		Point location = request.getLocation();
		((GraphicalEditPart)getHost()).getFigure().translateToRelative(location);
		EObject targetElement = (EObject)getHost().getModel();

		if(request instanceof CreateElementCompositionRequest) {
			CreateElementCompositionRequest req = (CreateElementCompositionRequest)request;
			return createElementCompositionCommand(req.getDragSource(), targetElement, location);
		}
		if(request instanceof CreateConnectionCompositionRequest) {
			CreateConnectionCompositionRequest req = (CreateConnectionCompositionRequest)request;
			return createConnectionCompositionCommand(req.getDragSource(), targetElement, location);
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public EditPart getTargetEditPart(Request request) {
		if(request instanceof CreateElementCompositionRequest ||
				request instanceof CreateConnectionCompositionRequest) {
			return getHost();
		}
		return super.getTargetEditPart(request);
	}

	/** Creates the relay command to {@link IConnectionCompositorService}. */
	private Command createConnectionCompositionCommand(ConnectionCompositionSource dragSource,
			final EObject targetElement, Point location) {

		final IConnectionCompositionContext context =
				createConnectionDropContext(getHost(), location);
		final EObject parentElement = dragSource.getDesignatedParentElement();
		final EObject droppedElement = dragSource.getDesignatedElement();

		return new Command() {
			@Override
			public boolean canExecute() {
				return IConnectionCompositorService.getInstance().canConnect(droppedElement,
						targetElement, parentElement, context);
			}

			@Override
			public void execute() {
				IConnectionCompositorService.getInstance().connect(droppedElement, targetElement,
						parentElement, context);
			}
		};
	}

	/** Creates the relay command to {@link IElementCompositorService}. */
	private Command createElementCompositionCommand(ElementCompositionSource dragSource,
			final EObject targetElement, Point location) {

		final EObject droppedElement = dragSource.getDesignatedElement();
		org.fortiss.tooling.base.model.layout.Point modelPoint =
				LayoutModelElementFactory.createPoint(location.x, location.y, "CompositionPoint");

		final IElementCompositionContext context;
		Control viewerControl = getHost().getRoot().getViewer().getControl();
		if(viewerControl instanceof FigureCanvas) {
			Point viewLocation = ((FigureCanvas)viewerControl).getViewport().getViewLocation();
			org.fortiss.tooling.base.model.layout.Point modelViewLocPoint =
					LayoutModelElementFactory.createPoint(viewLocation.x, viewLocation.y,
							"ViewLocationPoint");
			context = new ElementDropContext(targetElement, modelPoint, modelViewLocPoint,
					getHost() instanceof DiagramEditPartBase, true,
					ZoomUIUtils.determineZoom(getHost().getRoot()));
		} else {
			context = new ElementDropContext(targetElement, modelPoint,
					getHost() instanceof DiagramEditPartBase,
					ZoomUIUtils.determineZoom(getHost().getRoot()));
		}

		return new Command() {
			@Override
			public boolean canExecute() {
				return IElementCompositorService.getInstance().canCompose(targetElement,
						droppedElement, context);
			}

			@Override
			public void execute() {
				IElementCompositorService.getInstance().compose(targetElement, droppedElement,
						context);
			}
		};
	}
}
