/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static org.eclipse.draw2d.PositionConstants.EAST;
import static org.eclipse.draw2d.PositionConstants.SOUTH;
import static org.eclipse.draw2d.PositionConstants.SOUTH_EAST;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_CONNECTOR_SIZE;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_ASPECT_RATIO;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getConnectorAngleAsDouble;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;

import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.utils.EllipseLayoutUtils;
import org.fortiss.tooling.base.utils.LayoutModelElementFactory;

/**
 * Utility methods for ellipse-shaped edit part layout.
 * 
 * @author hoelzl
 */
public final class EllipseLayoutUIUtils {

	/**
	 * Returns Cartesian coordinates of the point on the ellipse border having
	 * the given angle.
	 * 
	 * @param angle
	 *            the angle
	 * @param ellipseSize
	 *            the wdith and height of the ellipse
	 * @param inset
	 *            the insets of the ellipse figure
	 * @return the point coordinates
	 */
	private static Point calculateEllipseBorderPoint(double angle, Dimension ellipseSize,
			int inset) {
		org.fortiss.tooling.base.model.layout.Dimension ellipseSizeElement =
				LayoutModelElementFactory.createDimension(ellipseSize.width(), ellipseSize.height(),
						ellipseSize.toString() + "_elem");
		org.fortiss.tooling.base.model.layout.Point cartesianPoint =
				EllipseLayoutUtils.calculateEllipseBorderPoint(angle, ellipseSizeElement, inset);
		return new Point(cartesianPoint.getX(), cartesianPoint.getY());
	}

	/**
	 * Returns a radian value for a given degree value
	 * 
	 * @param degree
	 *            the degree (0 to 360)
	 * @return rad (0 to 2*PI)
	 */
	public static double getRadianFromDegree(double degree) {
		return (Math.PI / 180.0) * degree;
	}

	/**
	 * Ensure that the minimum size and aspect ratio of a oval shape are
	 * redeemed.
	 * 
	 * @param oldSize
	 *            the old ellipse size
	 * @param direction
	 *            the resize direction
	 * @return the clamped shape size
	 */
	public static Dimension clampEllipseShapeSize(Dimension oldSize, int direction) {
		return clampEllipseShapeSize(oldSize, direction, DEFAULT_SHAPE_MINIMUM_WIDTH,
				DEFAULT_SHAPE_MINIMUM_HEIGHT, DEFAULT_SHAPE_ASPECT_RATIO);
	}

	/**
	 * Ensure that the minimum size and aspect ratio of a oval shape are
	 * redeemed.
	 * 
	 * @param oldSize
	 *            the old ellipse size
	 * @param direction
	 *            the resize direction
	 * @param minimumWidth
	 *            the minimal width
	 * @param minimumHeight
	 *            the minimal height
	 * @param aspectRatio
	 *            the aspect ratio
	 * @return the clamped shape size
	 */
	public static Dimension clampEllipseShapeSize(Dimension oldSize, int direction,
			int minimumWidth, int minimumHeight, double aspectRatio) {
		Dimension newSize;

		// check for minimum size
		newSize = new Dimension(oldSize.width < minimumWidth ? minimumWidth : oldSize.width,
				oldSize.height < minimumHeight ? minimumHeight : oldSize.height);

		// ensure that the aspect ratio is redeemed
		switch(direction) {
			case SOUTH_EAST:
				int aspectHeight = (int)(newSize.width / aspectRatio);
				int aspectWidth = (int)(newSize.height * aspectRatio);
				newSize.setSize(
						new Dimension(newSize.width > aspectWidth ? aspectWidth : newSize.width,
								newSize.height > aspectHeight ? aspectHeight : newSize.height));
				break;
			case SOUTH:
				newSize.setSize(new Dimension((int)(newSize.height * aspectRatio), newSize.height));
				break;
			case EAST:
				newSize.setSize(new Dimension(newSize.width, (int)(newSize.width / aspectRatio)));
				break;
		}

		return newSize;
	}

	/**
	 * Returns the bounds rectangle of the connector given its parent bounds.
	 * 
	 * @param connector
	 *            the connector element
	 * @param parent
	 *            the parent element
	 * @return the connector bounds
	 */
	public static Rectangle getAbsoluteConnectorBounds(ILayoutedModelElement connector,
			ILayoutedModelElement parent) {

		Rectangle result = new Rectangle(0, 0, DEFAULT_CONNECTOR_SIZE, DEFAULT_CONNECTOR_SIZE);
		double angle = getConnectorAngleAsDouble(connector);
		org.fortiss.tooling.base.model.layout.Dimension parentSize = getNodeSize(parent);

		Dimension parentDim = new Dimension(parentSize.getWidth(), parentSize.getHeight());
		Point position = calculateEllipseBorderPoint(angle, parentDim, DEFAULT_CONNECTOR_SIZE / 2)
				.translate(-DEFAULT_CONNECTOR_SIZE / 2, -DEFAULT_CONNECTOR_SIZE / 2);
		org.fortiss.tooling.base.model.layout.Point parentPos = getNodePosition(parent);
		result.x = position.x + parentPos.getX();
		result.y = position.y + parentPos.getY();

		return result;
	}
}
