/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static org.fortiss.tooling.base.utils.BaseMathUtils.convertNumber;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getEClassForClass;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.List;
import java.util.Random;
import java.util.function.BiConsumer;
import java.util.stream.DoubleStream;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.ToolingBaseActivator;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.AnnotationValueService;
import org.fortiss.tooling.base.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.annotation.valueprovider.AnnotationInstSpec;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationAction;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionEntry;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionEntry.ActionScope;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionInputParameters;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionParameters;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionReqInputParameter;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionReqInputParameters;
import org.fortiss.tooling.base.compose.HierarchicElementCompositorBase;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Utility methods for dealing with {@link IAnnotatedSpecification}s.
 * 
 * @author diewald
 */
public class AnnotationUtils {
	/**
	 * Returns the annotation of the particular {@code clazz} from a given {@code modelElement}.
	 */
	private static <T extends IAnnotatedSpecification> T
			getAnnotationInternal(final IModelElement modelElement, Class<T> clazz) {
		return pickFirstInstanceOf(clazz, modelElement.getSpecifications());
	}

	/**
	 * Creates an annotation of a particular {@code eClass} and adds it the specification list of
	 * the {@code modelElement}.
	 */
	@SuppressWarnings("unchecked")
	private static <T extends IAnnotatedSpecification> void
			createAnnotationInternal(final IModelElement modelElement, Class<T> clazz) {

		final EClass eClass = getEClassForClass(clazz);

		if(eClass == null) {
			return;
		}

		final T newAnnotation = (T)EcoreUtil.create(eClass);
		modelElement.getSpecifications().add(newAnnotation);
	}

	/**
	 * Returns the value of type {@code U} specified for the given {@link IModelElement} by the
	 * given annotation {@code clazz}.
	 */
	public static <T extends IAnnotatedSpecification, U> U
			getAnnotationValue(IModelElement modelElement, Class<T> clazz, Class<U> valueType) {
		IAnnotationValueProvider<T> annValProvider =
				getAnnotationValueProvider(modelElement, clazz);
		T annotation = getAnnotation(modelElement, clazz);
		U value = annValProvider.getAnnotationValue(annotation);
		if(value != null && !valueType.isAssignableFrom(value.getClass())) {
			throw new IllegalArgumentException("The value of the annotation " +
					clazz.getSimpleName() + " of the model element " + modelElement +
					" is not of the expected type " + valueType.getSimpleName());
		}
		return value;
	}

	/**
	 * Predicate if for the given {@link IModelElement}, an annotation of the given {@code type} has
	 * been registered.
	 */
	public static boolean hasAnnotation(IModelElement modelElement,
			Class<? extends IAnnotatedSpecification> type) {
		IAnnotationValueService as = IAnnotationValueService.getInstance();
		return as.getAnnotationEntry(modelElement).getSpecification(type) != null;
	}

	/**
	 * <p>
	 * Returns the annotation of a given {@code Class<T>} from a given {@code modelElement}. If the
	 * annotation does not exist yet, it is created and added to the specification list of the
	 * {@code modelElement}.
	 * </p>
	 * 
	 * <p>
	 * <b>Note: </b>The access to the model is wrapped into a command in order to allow for
	 * transparent creation of the annotation object from other contexts (e.g., GUI). If this is not
	 * desirable, use the {@link #getAnnotation(IModelElement, Class, boolean)} variant of this
	 * method.
	 * </p>
	 */
	public static <T extends IAnnotatedSpecification> T getAnnotation(IModelElement modelElement,
			Class<T> clazz) {
		return getAnnotation(modelElement, clazz, true);
	}

	/**
	 * Returns the annotation of a given {@code Class<T>} from a given {@code modelElement}. If the
	 * annotation does not exist yet, it is created and added to the specification list of the
	 * {@code modelElement}. Depending on the caller's choice, the model modification is wrapped
	 * into a command.
	 */
	public static <T extends IAnnotatedSpecification> T getAnnotation(
			final IModelElement modelElement, final Class<T> clazz, boolean wrapIntoCommand) {

		T annotation = getAnnotationInternal(modelElement, clazz);

		if(annotation != null) {
			return annotation;
		}

		if(wrapIntoCommand) {
			ITopLevelElement modelContext =
					IPersistencyService.getInstance().getTopLevelElementFor(modelElement);
			modelContext.runAsCommand(() -> {
				// TODO (see #2228 / #2208) Keep this code that ensures that non-existing
				// annotations are lazily instantiated until #2228 is resolved. It should be
				// noted that each time an annotation is constructed from this context, an
				// undo/redo command will be created that is not comprehensible for the user
				// (see #2208).
				createAnnotationInternal(modelElement, clazz);
			});
		} else {
			// Create if the creation does not need to be wrapped into a command.
			createAnnotationInternal(modelElement, clazz);
		}

		// Return new annotation (be it created in a command, or directly). This will return null in
		// case the creation of the new annotation was not successful.
		return getAnnotationInternal(modelElement, clazz);
	}

	/**
	 * Returns the {@link IAnnotationValueProvider} of the given {@code annotationType}.
	 * 
	 * @param element
	 *            {@link IModelElement} to which an annotation of the given type is bound (required
	 *            by the service).
	 * @param annotationType
	 *            Annotation type.
	 * @return The corresponding value provider or {@code null} if none is found.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends IAnnotatedSpecification> IAnnotationValueProvider<T>
			getAnnotationValueProvider(IModelElement element, Class<T> annotationType) {
		AnnotationEntry annEntry = AnnotationValueService.getInstance().getAnnotationEntry(element);
		return (IAnnotationValueProvider<T>)annEntry.getAnnotationValueProvider(annotationType);
	}

	/**
	 * Returns all {@link IAnnotatedSpecification}s attached to the given {@link IModelElement}.
	 * 
	 * @param modelElement
	 *            for which the attached {@link IAnnotatedSpecification}s shall be found.
	 * @return List of attached {@link IAnnotatedSpecification}s.
	 */
	public static List<IAnnotatedSpecification> getAnnotationsOf(IModelElement modelElement) {
		return pickInstanceOf(IAnnotatedSpecification.class, modelElement.getSpecifications());
	}

	/**
	 * Returns the {@link IAnnotatedSpecification}s matching the type 'T' of
	 * the parents of the given {@link IModelElement}.
	 * 
	 * @param startElement
	 *            The {@link IModelElement} to start the search from.
	 * @param clazz
	 *            The {@link IAnnotatedSpecification}s to be searched for.
	 * @return {@link EList} of found annotations.
	 */
	public static <T extends IAnnotatedSpecification> EList<T>
			getParentAnnotationsWithType(IModelElement startElement, Class<T> clazz) {
		EList<T> parentAnnotationList = new BasicEList<T>();
		List<IModelElement> parentsElements = getParentsWithType(startElement, IModelElement.class);

		for(IModelElement parentElement : parentsElements) {
			T annotation = pickFirstInstanceOf(clazz, parentElement.getSpecifications());
			parentAnnotationList.add(annotation);
		}
		return parentAnnotationList;
	}

	/**
	 * Creates an action that takes an average value and a bound to generate a random value. The
	 * action will be registered for the given {@link EStructuralFeatureDescriptor}. This
	 * method must be called from an {@link IAnnotationValueProvider} when the action list is built
	 * (typically {@link IAnnotationValueProvider#getContextMenuEntries(IAnnotatedSpecification)}).
	 * 
	 * Usually, this method is used to allow the creation of random values for displayed set of
	 * {@link IAnnotatedSpecification}s in the corresponding table from the context menu.
	 * 
	 * @param provider
	 *            {@link IAnnotationValueProvider} offering the ability to create random values.
	 * @param valType
	 *            {@link EStructuralFeatureDescriptor} describing the values' properties.
	 * @return {@link AnnotationAction}s to trigger random value generation.
	 */
	public static <T extends IAnnotatedSpecification> AnnotationAction<T> createRandValInputAction(
			IAnnotationValueProvider<T> provider, EStructuralFeatureDescriptor valType) {

		// Generates random entries around the given value and the fixed bound for all visible
		// annotations of the passed type.
		BiConsumer<AnnotationInstSpec<T>, AnnotationActionParameters<T>> randomValuePropagation =
				(AnnotationInstSpec<T> spec, AnnotationActionParameters<T> parameters) -> {
					T annotation = parameters.getAnnotation();

					EStructuralFeatureDescriptor featureDescriptor =
							provider.getEStructuralFeatureDescriptor();
					EClassifier featureType = featureDescriptor.getEType(annotation);

					AnnotationActionInputParameters params = parameters.getInputParameters();

					Object value = params.get("Average Value:", featureType.getInstanceClass());
					Object bound = params.get("Bound:        ", featureType.getInstanceClass());

					Double dValue = null;
					Double dBound = null;
					if(value instanceof Number && bound instanceof Number) {
						dValue = ((Number)value).doubleValue();
						dBound = ((Number)bound).doubleValue();
					}
					if(dValue == null || dBound == null) {
				// @CodeFormatterOff
						error(ToolingBaseActivator.getDefault(),
						      "Generation of random values for " + annotation.getName() + ": " +
						      "The average value or bound type is not number.");
						// @CodeFormatterOn
						return;
					}

					Random rand = new Random();
					DoubleStream doubleStream = rand.doubles(dValue - dBound, dValue + dBound);
					Number randVal = doubleStream.findAny().getAsDouble();

					// Number type is checked above already.
					@SuppressWarnings("unchecked") Class<Number> targetNumberType =
							(Class<Number>)featureType.getInstanceClass();
					randVal = convertNumber(randVal, targetNumberType);

					try {
						provider.setAnnotationValue(randVal, annotation);
					} catch(Exception e) {
				// @CodeFormatterOff
						error(ToolingBaseActivator.getDefault(),
						      "Could not set the value of the annotation " + annotation.getName() +
						      " when trying to propagate the value form the selected annotation.", e);
						// @CodeFormatterOn
					}
				};

		AnnotationActionReqInputParameters randInputParameters =
				new AnnotationActionReqInputParameters();
		randInputParameters.add(new AnnotationActionReqInputParameter("Average Value:", valType));
		randInputParameters.add(new AnnotationActionReqInputParameter("Bound:        ", valType));

		AnnotationActionEntry randValuePropDescriptor = new AnnotationActionEntry(
				"Create random values", ActionScope.ALL_VISIBLE_ITEMS, true, randInputParameters);

		return new AnnotationAction<T>(randValuePropDescriptor, randomValuePropagation);
	}

	/**
	 * Recursively instantiates the annotations for a given {@link IModelElement} and its offspring.
	 * This is useful to ensure that all annotations are available for programmatically created
	 * models. While the {@link HierarchicElementCompositorBase} could be used to ensure the
	 * instantiation of annotations in model-transformations, too, this is not always possible (or
	 * convenient).
	 * 
	 * @param element
	 *            {@link EObject} for which annotations should be instantiated recursively.
	 */
	public static void instantiateAnnotationsRecursive(EObject element) {
		IAnnotationValueService as = IAnnotationValueService.getInstance();
		if(element instanceof IModelElement) {
			as.instantiateAnnotations((IModelElement)element);
		}
		for(IModelElement child : getChildrenWithType(element, IModelElement.class)) {
			as.getAnnotationEntry(child);
		}

	}
}
